"""Utilities"""

from strand.taskrunning.base import Taskrunner
from strand.taskrunning.coroutine import CoroutineTaskrunner
from strand.taskrunning.multiprocess import MultiprocessTaskrunner
from strand.taskrunning.store_writer import StoreTaskWriter
from strand.taskrunning.thread import ThreadTaskrunner

from strand.constants import THREAD, PROCESS, SYNC, STORE, COROUTINE


def resolve_runner_cls(target):
    if isinstance(target, str):
        if target == THREAD:
            target_cls = ThreadTaskrunner
        elif target == PROCESS:
            target_cls = MultiprocessTaskrunner
        elif target == SYNC:
            target_cls = Taskrunner
        elif target == STORE:
            target_cls = StoreTaskWriter
        elif target == COROUTINE:
            target_cls = CoroutineTaskrunner
        else:
            raise ValueError(
                f'Taskrunner target {target} is invalid. '
                f'Valid targets are {THREAD}, {PROCESS}, {SYNC}, {STORE}, and {COROUTINE}.'
            )
    elif target == Taskrunner or issubclass(target, Taskrunner):
        target_cls = target
    else:
        raise ValueError(
            f'Taskrunner target {target} is invalid. Must be an allowed string or a Taskrunner class.'
        )
    return target_cls



"""A taskrunner that reads task definitions from a store and runs the tasks"""

from threading import Thread
from typing import Callable


class StoreTaskReader:
    _get_task_func: Callable

    def __init__(self, store, get_task_func=None):
        self._store = store
        self._get_task_func = get_task_func

    def _listen_handler(self):
        while True:
            available_tasks = list(self._store)
            # if available_tasks:
            #     for task in available_tasks:
            #         task_id = task['_id']
            #         if task['name'] in [
            #             'initiate_default_outlier_model',
            #             'continue_default_outlier_model',
            #         ]:
            #             continue
            #         func = available_task_types.get(task['name'], None)
            #         if not func:
            #             warn(
            #                 f'Task definition not found for task name {task["name"]}. Task cannot be run.'
            #             )
            #             update = {
            #                 'task_status': 'error',
            #                 'error': 'Task definition not found',
            #             }
            #         else:
            #             print(f'Running task: {str(task)}')
            #             task_manager.update_task_by_id(
            #                 task_id, {'task_status': 'running'}
            #             )
            #             try:
            #                 args = task['args']
            #                 kwargs = task['kwargs']
            #                 result = func(*args, **kwargs)
            #                 print(f'Task success: {not result.get("error", False)}')
            #                 update = {'task_status': 'complete'}
            #             except Exception:
            #                 error_trace = traceback.format_exc()
            #                 traceback.print_exc()
            #                 update = {
            #                     'task_status': 'error',
            #                     'error_trace': error_trace,
            #                 }
            #         task_manager.update_task_by_id(task_id, update)

    def listen(self):
        t = Thread(target=self._listen_handler)
        t.start()



"""Taskrunner base class"""

from typing import Callable, Optional


class Taskrunner:
    _on_iter: Optional[Callable]
    _on_end: Optional[Callable]
    _on_error: Optional[Callable]

    def __init__(
        self,
        func: Callable,
        *args,
        on_iter: Optional[Callable] = None,
        on_end: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
        **kwargs
    ):
        self._func = func
        self._on_iter = on_iter
        self._on_end = on_end
        self._on_error = on_error
        self._init_args = args
        self._init_kwargs = kwargs

    def __call__(self, *args, **kwargs):
        try:
            result = self._func(*self._init_args, *args, **self._init_kwargs, **kwargs)
            if self._on_iter:
                for value in result:
                    self._on_iter(value)
            if self._on_end:
                self._on_end(result)
            return result
        except Exception as err:
            if self._on_error:
                self._on_error(err)
            else:
                raise err



"""A taskrunner that writes task definitions to a store"""

from time import time

import dill
import nanoid
from .base import Taskrunner
from .store_reader import StoreTaskReader


class StoreTaskWriter(Taskrunner):
    task_id: str

    def __init__(
        self,
        func,
        store,
        *args,
        read_store=None,
        pickle_func=False,
        get_result=None,
        **kwargs
    ):
        Taskrunner.__init__(self, func, *args, **kwargs)
        self._store = store
        self._get_result = get_result
        self._pickle_func = pickle_func
        if read_store:
            StoreTaskReader(store=read_store).listen()

    @property
    def result(self):
        if not self._task_id:
            return None
        if self._get_result:
            task_result = self._get_result(self._task_id)
        else:
            task_dict = self._store[self._task_id]
            task_result = task_dict.get('result', None)
        return task_result

    def __call__(self, *args, task_id: str = None, **kwargs):
        task_dict = {
            'name': self._func.__name__,
            'func': None,
            'init_args': list(self._init_args),
            'init_kwargs': dict(self._init_kwargs),
            'args': list(args),
            'kwargs': dict(kwargs),
            'start_time': int(time() * 1000),
            'task_status': 'new',
        }
        if self._pickle_func:
            task_dict['func'] = dill(self._func)
        if not task_id:
            task_id = nanoid.generate()
        self._task_id = task_id
        self._store[task_id] = task_dict



"""Defines the TaskDispatcher class."""

import time
from typing import Callable, Optional

from strand.taskrunning.base import Taskrunner
from strand.taskrunning.utils import resolve_runner_cls
from strand.constants import THREAD


class TaskDispatcher:
    """A class for dispatching multiple taskrunners and collecting their outputs.

    Use the dispatch method to launch a task. Dispatch as many tasks as you want
    with the same dispatcher, then check the property dispatcher.complete if you need
    to confirm that all of the tasks have finished executing.

    Results of tasks can be found in the 'results', 'status', and 'errors' properties.

    >>> from strand import TaskDispatcher
    >>> import requests
    >>> from time import sleep

    >>> dispatcher = TaskDispatcher()
    >>> _ = dispatcher.dispatch(requests.get, args=['https://zombo.com'])
    >>> dispatcher.status
    [None]
    >>> dispatcher.errors
    [None]
    >>> dispatcher.results
    [None]
    >>> sleep(1)
    >>> dispatcher.status
    ['success']
    >>> dispatcher.errors
    [None]
    >>> dispatcher.results
    [<Response [200]>]
    """

    _is_complete = False
    results = []
    status = []
    errors = []
    _queued_runners = []
    _dispatched_runners = []
    _target_cls = Taskrunner

    def __init__(self, target=THREAD, check_complete=None):
        """
        Initialize the dispatcher and set configuration.

        :param target: Optional - the default taskrunner target (a taskrunner class or a string
            ('thread', 'process', 'coroutine', or 'store'). Defaults to 'thread'.
        :param check_complete: Optional - a callable that evaluates whether all of the dispatcher's
            tasks are complete.
        """
        self._target_cls = resolve_runner_cls(target)
        self._check_complete = check_complete

    def enqueue(
        self,
        func: Callable,
        *init_args,
        target=None,
        on_iter: Optional[Callable] = None,
        on_end: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
        args=None,
        kwargs=None,
        **init_kwargs
    ):
        self._is_complete = False
        if not kwargs:
            kwargs = {}
        target_cls = resolve_runner_cls(target) if target else self._target_cls
        runner = target_cls(
            func,
            *init_args,
            on_iter=on_iter,
            on_end=self._mk_end_handler(on_end),
            on_error=self._mk_error_handler(on_error),
            **init_kwargs
        )
        self._queued_runners.append((runner, args, kwargs))

    def dispatch(
        self,
        func,
        *init_args,
        target=None,
        on_iter: Optional[Callable] = None,
        on_end: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
        args=None,
        kwargs=None,
        **init_kwargs
    ):
        """Execute a single task. The results of the task will be appended to
        the results, status, and errors lists.

        """
        self._is_complete = False
        if not kwargs:
            kwargs = {}
        target_cls = resolve_runner_cls(target) if target else self._target_cls
        runner = target_cls(
            func,
            *init_args,
            on_iter=on_iter,
            on_end=self._mk_end_handler(on_end),
            on_error=self._mk_error_handler(on_error),
            **init_kwargs
        )
        self._dispatched_runners.append(runner)
        return runner(*args, **kwargs)

    def run(self, timeout=10):
        """Execute all enqueued tasks.

        """
        while len(self._queued_runners):
            runner, args, kwargs = self._queued_runners.pop()
            self._dispatched_runners.append(runner)
            runner(*args, **kwargs)
        end_time = time.time() + timeout
        while time.time() < end_time:
            if self.complete:
                return True
            time.sleep(0.01)
        raise TimeoutError()

    def _mk_end_handler(self, on_end):
        target_index = len(self.results)
        self.results.append(None)
        self.status.append(None)

        def end_handler(result):
            self.results[target_index] = result
            self.status[target_index] = 'success'
            if on_end:
                on_end(result)

        return end_handler

    def _mk_error_handler(self, on_error):
        target_index = len(self.errors)
        self.errors.append(None)

        def error_handler(result):
            self.errors[target_index] = result
            self.status[target_index] = 'error'
            if on_error:
                on_error(result)

        return error_handler

    def reset(self):
        self._is_complete = False
        self.results = []
        self.status = []
        self.errors = []
        self._queued_runners = []
        self._dispatched_runners = []

    @property
    def complete(self):
        if self._is_complete:
            return True
        if self._check_complete:
            self._is_complete = self._check_complete(
                self.results, self.status, self.errors,
            )
        else:
            for status in self.status:
                if not status:
                    return False
            self._is_complete = True
        return self._is_complete



"""Taskrunners"""

from .base import Taskrunner
from .coroutine import CoroutineTaskrunner
from .decorators import as_task
from .multi_dispatch import TaskDispatcher
from .multiprocess import MultiprocessTaskrunner
from .store_writer import StoreTaskWriter
from .store_reader import StoreTaskReader
from .thread import ThreadTaskrunner
